// FnSocket.cpp : implementation file
//

#include "stdafx.h"
#include "FnSocket.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define BUFFERSIZE	0x7fff

///////////////////////////////////////////////////
// CBuffer implementation
 
// CTcpIpBuffer constructor
CFnSocket::CBuffer::CBuffer()
{
	m_buflen=0;
	m_buflenmax=BUFFERSIZE;
	m_pIp=NULL;
	m_buf=new BYTE[m_buflenmax+1];
	m_buf[0]=0;
}

// CTcpIpBuffer denstructor
CFnSocket::CBuffer::~CBuffer()
{
	delete m_buf;
}

// Buffer realocation
BOOL CFnSocket::CBuffer::Realloc(UINT nsize)
{
	LPCSTR fname="Realloc(UINT)";
	BOOL retv=TRUE;

	m_crs.Lock();
	BYTE *tmp=(BYTE *)realloc(m_buf,nsize);
	if (tmp) {
		m_buf=tmp;
		m_buflenmax=nsize;
	} else {
		m_pIp->OnError("Memory Allocation Error in CFnSocket::CBuffer::Realloc()");
		retv=FALSE;
	}

	m_crs.Unlock();
	return retv;
}

// Add byte to buffer
BOOL CFnSocket::CBuffer::Add(char b)
{
	// add byte to buffer
	m_crs.Lock();
	m_buf[m_buflen++]=b;
	m_crs.Unlock();

	return TRUE;
}

// Add string to buffer
BOOL CFnSocket::CBuffer::Add(LPCSTR s)
{
	return Add((void *)s,strlen(s));
}

// Add binary data to buffer
BOOL CFnSocket::CBuffer::Add(LPCVOID m, UINT ln)
{
	// add binary data to buffer
	if (m_buflen+ln+1>m_buflenmax) Realloc(m_buflen+ln+1);
	m_crs.Lock();
	memcpy(&m_buf[m_buflen],m,ln);
	m_buflen+=ln; m_buf[m_buflen]=0;
	m_crs.Unlock();

	return TRUE;
}

// delete n bytes from buffer
BOOL CFnSocket::CBuffer::Del(UINT n)
{
	if (n>m_buflen) n=m_buflen;
	m_crs.Lock();
	for (UINT i=0; i<m_buflen-n; i++) m_buf[i]=m_buf[i+n];
	m_buflen-=n; m_buf[m_buflen]=0;
	m_crs.Unlock();
	return TRUE;
}

// Purge any data from buffer ans realloc memory
void CFnSocket::CBuffer::Purge()
{
	m_crs.Lock();
	m_buflen=0;
	m_crs.Unlock();
	Realloc(BUFFERSIZE);
}

/////////////////////////////////////////////////////////////////////////////
// CFnSocket

CFnSocket::CFnSocket()
{
	m_bConnectInProcess=FALSE;
	m_bConnected=FALSE;
	m_bTheadActive=FALSE;
	m_Stats.m_bytes_in=0;
	m_Stats.m_bytes_out=0;
	m_pRSock=NULL;
	m_Wnd.m_pFnSocket=this;
	if (!AfxSocketInit()) OnError("AfxSocketInit() failed");
	AfxBeginThread((AFX_THREADPROC)WndThreadProc,this,THREAD_PRIORITY_HIGHEST);
}

CFnSocket::~CFnSocket()
{
	m_bConnected=FALSE;
	m_Wnd.SendMessage(WM_FNSOCKET_CLOSE);
	m_Wnd.EndModalLoop(0);
	while (IsWindow(m_Wnd.m_hWnd)) Sleep(1);
	m_Wnd.Detach();
}


// Do not edit the following lines, which are needed by ClassWizard.
#if 0
BEGIN_MESSAGE_MAP(CFnSocket, CAsyncSocket)
	//{{AFX_MSG_MAP(CFnSocket)
	ON_MESSAGE(WM_SOCKET_NOTIFY, OnSocketNotify)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()
#endif	// 0

/////////////////////////////////////////////////////////////////////////////
// CFnSocket member functions

void CFnSocket::OnAccept(int nErrorCode) 
{
	// TODO: Add your specialized code here and/or call the base class
	if (!nErrorCode) {
		if (Accept(*m_pRSock)) {
			m_pRSock->m_bConnected=TRUE;
			m_pRSock->m_Wnd.m_bCreated=TRUE;
			m_InBuf.Purge();
			m_OutBuf.Purge();
			AfxBeginThread(FnPutThreadProc,m_pRSock);
			m_pRSock->OnConnectFin(0);
		} else {
			OnError("Error in CAsyncSocket::Accept()");
			m_pRSock->m_bConnected=FALSE;
		}
	} else {
		OnError("Error in CFnSocket::OnAccept()");
		m_bConnected=FALSE;
	}
	CAsyncSocket::OnAccept(nErrorCode);
}

void CFnSocket::OnClose(int nErrorCode) 
{
	// TODO: Add your specialized code here and/or call the base class
	m_bConnected=FALSE;
	while (m_bTheadActive) Sleep(1);
	CAsyncSocket::OnClose(nErrorCode);
	CAsyncSocket::Close();
	OnDisconnect(nErrorCode);
}


void CFnSocket::OnConnect(int nErrorCode) 
{
	// TODO: Add your specialized code here and/or call the base class
	if (nErrorCode) OnError("Error in CFnSocket::OnConnect()");
	if (!nErrorCode) {
		m_bConnectInProcess=TRUE;
		m_InBuf.Purge();
		m_OutBuf.Purge();
		if (m_Wnd.m_bCreated) {
			m_bConnected=TRUE;
			AfxBeginThread(FnPutThreadProc,this);
			OnConnectFin(nErrorCode);
		} else m_bConnected=FALSE;
		m_bConnectInProcess=FALSE;
	} else m_bConnected=FALSE;
	CAsyncSocket::OnConnect(nErrorCode);
}

void CFnSocket::OnReceive(int nErrorCode) 
{
	if (!nErrorCode) {
		DWORD rbytes=0L;
		if (IOCtl(FIONREAD,&rbytes)) {
			char *buf=new char[rbytes+1];
			Receive(buf,rbytes);
			buf[rbytes]=0;
			m_InBuf.Add(buf,rbytes);
			m_Stats.m_bytes_in+=rbytes;
			delete buf;
		} else OnError("Error in CAsyncSocket::IOCtl()");
	}
}

void CFnSocket::OnSend(int nErrorCode) 
{
	m_bConnected=TRUE;
}

static UINT FnErrorThreadProc(LPVOID param)
{
	CString errmsg=*(CString *)param;
	delete param;
	Beep(5000,50);
	AfxMessageBox(errmsg,MB_OK|MB_ICONWARNING|MB_SYSTEMMODAL);
	return 0;
}

void CFnSocket::OnError(LPCSTR pszComment)
{
	CString msg,comment;
	int code=GetLastError();
	if (code) msg.Format("Windows Sockets error code # %d\n\n",code);
	if (pszComment) comment=pszComment;
	if (msg.GetLength()>0 || comment.GetLength()>0) {
#ifdef _DEBUG
		AfxBeginThread(FnErrorThreadProc,(LPVOID)new CString(msg+comment));
#endif
	}
}

BOOL CFnSocket::Connect( LPCTSTR lpszHostAddress,UINT nHostPort,BOOL bClearBuffers)
{
	if (IsConnected() && m_hSocket!=INVALID_SOCKET) Disconnect();
	if (bClearBuffers) ClearIOBuffers();
	while (!IsWindow(m_Wnd.m_hWnd)) Sleep(1);
	m_Wnd.SendMessage(WM_FNSOCKET_CREATE,0);		// call Create()
	BOOL one=TRUE;
	SetSockOpt(SO_REUSEADDR,(char *)&one, sizeof(one));
	BOOL retv=CAsyncSocket::Connect(lpszHostAddress,nHostPort);
	int errcode=GetLastError();
	if (!retv && errcode!=WSAEWOULDBLOCK) {
		CString cs;
		cs.Format("Error in CFnSocket::Connect()  GetLastError()=%d",errcode);
		OnError(cs);
	}
	return retv;
}

BOOL CFnSocket::Connect(const SOCKADDR* lpSockAddr,int nSockAddrLen,BOOL bClearBuffers)
{
	if (IsConnected() && m_hSocket!=INVALID_SOCKET) Disconnect();
	if (bClearBuffers) ClearIOBuffers();
	while (!IsWindow(m_Wnd.m_hWnd)) Sleep(1);
	m_Wnd.SendMessage(WM_FNSOCKET_CREATE,0);		// call Create()
	BOOL one=TRUE;
	SetSockOpt(SO_REUSEADDR,(char *)&one, sizeof(one));
	BOOL retv=CAsyncSocket::Connect(lpSockAddr,nSockAddrLen);
	int errcode=GetLastError();
	if (!retv && errcode!=WSAEWOULDBLOCK) {
		CString cs;
		cs.Format("Error in CFnSocket::Connect()  GetLastError()=%d",errcode);
		OnError(cs);
	}
	return retv;
}


void CFnSocket::Close()
{
	if (m_hSocket!=INVALID_SOCKET) {
		CAsyncSocket::Close();
		m_Wnd.m_bCreated=FALSE;
		m_hSocket=INVALID_SOCKET;
	}	
}

void CFnSocket::Disconnect()
{
	if (m_bConnectInProcess) {
		while (m_bConnectInProcess) Sleep(1);
		Sleep(100);
	}
	m_bConnected=FALSE;
	while (m_bTheadActive) Sleep(1);
	m_Wnd.SendMessage(WM_FNSOCKET_CLOSE);
}

void CFnSocket::FnDisconnect()
{
	Close();
}

BOOL CFnSocket::Get(void *buffer,UINT length)
{
	UINT retv=FALSE;
	if (buffer && length<=m_InBuf.Len()) {
		m_InBuf.Lock();
		memcpy(buffer,m_InBuf.Buf(),length);
		m_InBuf.Unlock();
		m_InBuf.Del(length);
		retv=TRUE;
	}
	return retv;
}

static UINT FnPutThreadProc(LPVOID param)
{
	CFnSocket& s=*(CFnSocket *)param;
	s.m_bTheadActive=TRUE;
	while (s.IsConnected()) {
		UINT ln=s.m_OutBuf.Len();
		if (ln>0) {
			UINT suc=s.Send((void *)s.m_OutBuf.Buf(),ln);
			if (suc!=SOCKET_ERROR) {
				s.m_OutBuf.Del(suc);
				s.m_Stats.m_bytes_out+=suc;
			} else {
				if (GetLastError()!=WSAEWOULDBLOCK) {
					s.OnError("Error in CAsyncSocket::Send()");
					break;
				}

			}
		} else Sleep(1);
	}
	s.m_bTheadActive=FALSE;
	return 0;
}

BOOL CFnSocket::Put(LPCVOID buffer,UINT length)
{
	UINT retv=FALSE;
	if (buffer) {
		m_OutBuf.Add(buffer,length);
		retv=TRUE;
	}
	return retv;
}

BOOL CFnSocket::Gets(CString& str)
{
	BOOL retv=FALSE;
	UINT len=0;

	str="";
	m_InBuf.Lock();
	char *pBuf=(char *)m_InBuf.Buf();
	char *pNl=NULL;
	pNl=strchr(pBuf,'\r');
	if (!pNl) pNl=strchr(pBuf,'\n');
	if (!pNl && !IsConnected()) pNl=pBuf+m_InBuf.Len();
	if (pNl) {
		len=pNl-pBuf;
		for (UINT i=0; i<len; i++) {
			char c=pBuf[i];
			switch (c) {
			case '\r': break;
			case '\n': break;
			case '\b': 
				{
					int strln=str.GetLength();
					if (strln>0) str=str.Left(strln-1);
					break;
				}
			default: str+=pBuf[i]; 
			}
		}
		if (len>0) retv=TRUE;
	}
	m_InBuf.Unlock();
	m_InBuf.Del(len);
	if (*pBuf=='\r') m_InBuf.Del(1);
	if (*pBuf=='\n') m_InBuf.Del(1);
	return retv;
}

BOOL CFnSocket::Puts(CString str)
{
	Put((void *)(LPCSTR)str,str.GetLength());
	m_OutBuf.Add('\n');
	return TRUE;
}

BOOL CFnSocket::Putf(LPCSTR fmt,...)
{
	CString tmp;
	va_list pvar;

	va_start(pvar,fmt);
	tmp.FormatV(fmt,pvar);
	va_end(pvar);
	
	return Put((void *)(LPCSTR)tmp,tmp.GetLength());
}

void CFnSocket::ClearStatistics()
{
	m_Stats.m_bytes_in=0;
	m_Stats.m_bytes_out=0;
}

void CFnSocket::GetStatistics(DWORD* pIn,DWORD* pOut)
{
	*pIn=m_Stats.m_bytes_in;
	*pOut=m_Stats.m_bytes_out;
}

BOOL CFnSocket::Server(UINT port,CFnSocket *rsock)
{
	BOOL retv=FALSE;
	while (!IsWindow(m_Wnd.m_hWnd)) Sleep(1);
	m_Wnd.SendMessage(WM_FNSOCKET_CREATE,(WPARAM)port);		// call Create
	BOOL one=TRUE;
	SetSockOpt(SO_REUSEADDR,(char *)&one, sizeof(one));
	m_pRSock=rsock;
	if (!Listen()) {
		OnError("Error in CAsyncSocket::Listen()");
	} else retv=TRUE;
	return retv;
}

void CFnSocket::ClearIOBuffers()
{
	m_InBuf.Purge();
	m_OutBuf.Purge();
}

int CFnSocket::Send(const void* lpBuf, int nBufLen,int nFlags)
{
	return CAsyncSocket::Send(lpBuf,nBufLen,nFlags);
}

int CFnSocket::Receive(void* lpBuf,int nBufLen,int nFlags)
{
	return CAsyncSocket::Receive(lpBuf,nBufLen,nFlags);
}

void CFnSocket::WndThreadProc(CFnSocket &sock)
{
	WSADATA wsa;
	AfxSocketInit(&wsa);
	TRACE("WndThreadProc started\n");
	sock.m_Wnd.Create(NULL,"FnSocket Window",WS_OVERLAPPEDWINDOW);
	sock.m_Wnd.BeginModalState();
	sock.m_Wnd.RunModalLoop();
	sock.m_Wnd.EndModalState();
	sock.m_Wnd.Detach();
	TRACE("WndThreadProc exited\n");
}
