#if !defined(AFX_FNSOCKET_H__91989C25_C8A9_4987_B396_DDC38E9B2176__INCLUDED_)
#define AFX_FNSOCKET_H__91989C25_C8A9_4987_B396_DDC38E9B2176__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
// FnSocket.h : header file
//

#include <afxsock.h>
#include <afxmt.h>
#include "FnSocketWnd.h"

/////////////////////////////////////////////////////////////////////////////
// CFnSocket command target

class CFnSocket : public CAsyncSocket
{
	friend class CFnSocketWnd;
public:
	// Internal input and output queue
	class CBuffer {
	public:
		CCriticalSection m_crs;
		BYTE *m_buf;
		UINT m_buflen;
		UINT m_buflenmax;
		CFnSocket *m_pIp;
	public:
		CBuffer();
		~CBuffer();
		// member functions
		const BYTE * Buf() {return m_buf;}	// buffer accecc
		UINT Len() {return m_buflen;}		// data length
		BOOL Realloc(UINT nsize);			// buffer realloc
		// sync functions
		void Lock() {m_crs.Lock();}
		void Unlock() {m_crs.Unlock();}
		// data access functions
		BOOL Add(char c);					// add single character
		BOOL Add(LPCSTR s);					// add ASCIIZ string
		BOOL Add(LPCVOID m,UINT ln);			// add binary data
		BOOL Del(UINT ln);					// delete and drag tail
		BYTE & operator[](int i) {return m_buf[i];}
		void Purge();
	} m_InBuf,m_OutBuf;

	// statistics
	struct CTcpIpStatistics {
		DWORD m_bytes_in;
		DWORD m_bytes_out;
	} m_Stats;

public:
	static void WndThreadProc(CFnSocket &sock);
	CFnSocketWnd m_Wnd;

// Creation
	CFnSocket();
	virtual ~CFnSocket();

// Information
	virtual BOOL IsConnected() {return m_bConnected;}
	virtual void GetStatistics(DWORD* pIn,DWORD* pOut);

// Non-blocking operations
	virtual BOOL Connect(LPCTSTR lpszHostAddress,UINT nHostPort,BOOL bClearBuffers=TRUE);
	virtual BOOL Connect(const SOCKADDR* lpSockAddr,int nSockAddrLen,BOOL bClearBuffers=TRUE);
	virtual BOOL Get(void *buffer,UINT length);
	virtual BOOL Gets(CString& str);
	virtual BOOL Put(LPCVOID buffer,UINT length);
	virtual BOOL Puts(CString str);
	virtual BOOL Putf(LPCSTR fmt,...);
	virtual void Disconnect();
	virtual BOOL Server(UINT port,CFnSocket *rsock);
	virtual void ClearStatistics();
	virtual void ClearIOBuffers();
	
// Blocking operations


// Overridable Notification Functions

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CFnSocket)
	virtual void OnDisconnect(int nErrorCode)=0;
	virtual void OnConnectFin(int nErrorCode)=0;
	virtual void OnError(LPCSTR pszComment);
	//}}AFX_VIRTUAL

	// Generated message map functions
	//{{AFX_MSG_MAP(CFnSocket)
		// NOTE - the ClassWizard will add and remove member functions here.
	//}}AFX_MSG

protected:
	BOOL m_bConnected;
	BOOL m_bTheadActive;
	CFnSocket *m_pRSock;
	virtual void Close();
	virtual void OnConnect(int nErrorCode);
	virtual void OnAccept(int nErrorCode);
	virtual void OnClose(int nErrorCode);
	virtual void OnReceive(int nErrorCode);
	virtual void OnSend(int nErrorCode);
	virtual int Send(const void* lpBuf, int nBufLen,int nFlags=0);
	virtual int Receive(void* lpBuf,int nBufLen,int nFlags=0);

	friend class CBuffer;
	friend UINT FnPutThreadProc(LPVOID param);
	void FnDisconnect();
private:
	BOOL volatile m_bConnectInProcess;
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_FNSOCKET_H__91989C25_C8A9_4987_B396_DDC38E9B2176__INCLUDED_)
