// WinSTDP.cpp: implementation of the CWinSTDP class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "WinSTDP.h"
#include "stdp_ie.h"
#include "TGW6.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWinSTDP::CWinSTDP()
{
	m_bReconnectInProgress=FALSE;
	m_lrecvtime=GetTickCount();
	m_lsendtime=GetTickCount();
	m_stat=CONN_NOT_SET;
}

CWinSTDP::~CWinSTDP()
{

}

void CWinSTDP::OnDisconnect(int nErrorCode)
{
	PrintLogWin(m_ncit,"OnDisconnect():    STDP ");
	TRACE("CWinSTDP::OnDisconnect() called\n");
	m_CrsOut.Lock();
	for(int i=0; i<m_OutQueue.GetSize(); i++) {

		CMessage & msg = m_OutQueue[i];
		if(msg.GetSendTime()==0) continue;
		if(msg.GetAck()) {
			m_OutQueue.RemoveAt(i--);
			continue;
		}
		msg.m_sendtime=0; // try to send this message later
	}
	m_CrsOut.Unlock();

	switch(m_stat) {
		case CONN_HIDER_ERR:
		case CONN_UNK_CLIENT:
		case CONN_UNK_SERVER:
        case CONN_ACCESS:
		case CONN_CORRUPT:
			PrintLogWin(m_ncit,"OnDisconnect():  STDP      ");
			break;
        case CONN_REP_CONNECT:
			PrintLogWin(m_ncit,"OnDisconnect():  STDP -    , ");
			Reconnect();
			TRACE("OnDisconnect() CONN_REP_CONNECT\n");
			break;
		case CONN_OK:
			m_stat=CONN_NOT_SET;
        case CONN_NOT_SET:
			PrintLogWin(m_ncit,"OnDisconnect():  STDP -   , ");
			Reconnect();
			TRACE("OnDisconnect() CONN_NOT_SET\n");
			break;
        case CONN_CHAN_CLOSED:
			PrintLogWin(m_ncit,"OnDisconnect():  STDP -   , ");
			Reconnect();
			TRACE("OnDisconnect() CONN_CHAN_CLOSED\n");
			break;
		default:
			break;
	}
}

void CWinSTDP::OnConnectFin(int nErrorCode) 
{
	TRACE("CWinSTDP::OnConnectFin() nErrorCode=%d m_stat=%d\n",nErrorCode,m_stat);
	if(nErrorCode==0) {
		PrintLogWin(m_ncit,"OnConnectFin():    TCP/IP,  STDP");
		m_stat=CONN_NOT_SET;
		AfxBeginThread((AFX_THREADPROC)Scheduler,(LPVOID)this);
	} else {
		PrintLogWin(m_ncit,"OnConnectFin(): o      TCP/IP");
		m_stat=CONN_NOT_SET;
	}
}

void CWinSTDP::OnError(LPCSTR pszComment)
{
	TRACE("CWinSTDP::OnError() m_stat=%d pszComment='%s'\n",m_stat,pszComment);
	PrintLogWin(m_ncit,"OnError():    TCP/IP,  ");
	Reconnect();
}

void CWinSTDP::Scheduler(CWinSTDP * pSTDP)
{
	TRACE("CWinSTDP::Scheduler() started\n");
	pSTDP->m_lrecvtime=GetTickCount();
	pSTDP->m_lsendtime=GetTickCount();
	//    login-a
	pSTDP->PrintLogWin(pSTDP->m_ncit,"Scheduler():   ...");
	pSTDP->PrintLogWin(pSTDP->m_ncit,"Scheduler():     STDP ");
    pSTDP->Putf("STDP0480000000003%s%s00100",pSTDP->m_laa,pSTDP->m_raa);

	while (pSTDP->IsConnected() || pSTDP->GetIUType()>=0) {
		pSTDP->SendTestIU();
		pSTDP->SendIU();
		pSTDP->ParseIU();
		pSTDP->CheckQueue();
		pSTDP->CheckTimeout();
		Sleep(1);
	}
	TRACE("CWinSTDP::Scheduler() ended  pSTDP->m_stat=%d\n",pSTDP->m_stat);
	if (!pSTDP->IsConnected()) 
		pSTDP->Reconnect();
	pSTDP->PrintLogWin(pSTDP->m_ncit,"Scheduler():   ...");
}

void CWinSTDP::Connect(LPCSTR ip, UINT port, LPCSTR raa, LPCSTR laa,int nconf)
{
	m_rip=ip;
	m_port=port;
	m_raa=raa;
	m_laa=laa;
	m_nconf=nconf;
	CFnSocket::Connect(m_rip,m_port);
	PrintLogWin(m_ncit,"Connect():    c   TCP/IP");
	TRACE("CWinSTDP::Connect() called\n");
}

UINT CWinSTDP::Hex2Int(LPCSTR str, UINT len)
{
   LPCSTR tab="0123456789ABCDEF";
   int i=1;
   UINT sum = 0;
   
   if(strlen(str)<len) {len = strlen(str)-1;}
    while(len--){
        char * x=strchr(tab,str[len]);
        if (x==NULL) return sum;
        sum+=((int)(x-tab))*i;
        i*=16;
    }
    return(sum);
}

UINT CWinSTDP::Dec2Int(LPCSTR str, UINT len)
{
	CString cs(str,len);
	return (UINT)atol(cs);
}

void CWinSTDP::TransMem(void *mem,const void * tab, UINT len)
{
    LPBYTE m=(LPBYTE)mem;
    LPBYTE t=(LPBYTE)tab;
    for(UINT i=0; i<len; i++) m[i]=t[m[i]];
}

int CWinSTDP::Send(const void* lpBuf, int nBufLen,int nFlags)
{
	char * buf= new char [nBufLen];
	memcpy(buf,lpBuf,nBufLen);
	TransMem(buf,win2koi,nBufLen);
	TransMem(buf,KOI_TO_DKOI,nBufLen);
	int rv=CFnSocket::Send(buf,nBufLen,nFlags);
	m_lsendtime=GetTickCount();
	delete buf;
	return rv;	
}

int CWinSTDP::Receive(void* lpBuf,int nBufLen,int nFlags)
{
	int rv=CFnSocket::Receive(lpBuf,nBufLen,nFlags);
	if (rv!=SOCKET_ERROR) {
		TransMem(lpBuf,POST_DKOI_ASCII,rv);
		TransMem(lpBuf,koi2win,rv);
		m_lrecvtime=GetTickCount();
	}
	return rv;
}

int CWinSTDP::GetIUType(void) {
    unsigned int bsize=m_InBuf.Len();      //     e
    
    if(bsize<sizeof(FixHider)) return -1;  //    

	FixHider * fh = (FixHider *)m_InBuf.m_buf;
    if(memcmp(fh->stdp,"STDP",4)!=0) return -2;
    /*
           STDP
        IU 
     */
    if(bsize<GetIUSize()) return -1;

    return Dec2Int((LPCSTR)fh->type,sizeof(fh->type));
}

unsigned CWinSTDP::GetIUSize(void)
{
    FixHider fh;
    memcpy(&fh,m_InBuf.m_buf,sizeof(fh));
    return Dec2Int((LPCSTR)fh.h_len,3)+Dec2Int((LPCSTR)fh.v_len,8);
}

unsigned CWinSTDP::GetMessageSize(void)
{
    MsgHider fh;
    memcpy(&fh,m_InBuf.m_buf,sizeof(fh));
    return Dec2Int((LPCSTR)fh.msg_len,8);
}


void CWinSTDP::ParseIU()
{
	switch(m_stat) {
		case CONN_OK:
			switch(GetIUType()) {
			case -2: m_stat=CONN_CORRUPT; m_InBuf.Purge(); 
				PrintLogWin(m_ncit,"ParseIU():  STDP -     ");
				break;
			case -1: break;
			case IE_TST:
				{
					struct TestHider * th=(TestHider *)m_InBuf.m_buf;
					int cc=Hex2Int((char *)th->ctl,2);
					Putf("STDP0470000000009%s%s%02x00",m_laa,m_raa,cc);
					PrintLogWin(m_ncit,"ParseIU():    ,  ");
					TRACE("IE_TEST Received\n");
				}
			case IE_MSGANSW:
				{
					//    
					MsgAnswHider * ca=(MsgAnswHider *)m_InBuf.m_buf;
					int rv=(CWinSTDP::WinSTDPStatus)Hex2Int((char *)ca->code,2);
					CMessage * pMsg=FindMessage((LPCSTR)ca->sndr_num);
					if(pMsg!=NULL) pMsg->SetAck(TRUE);
					DelIU();
					PrintLogWin(m_ncit,"ParseIU():      ");
					TRACE("IE_MSGANSW Received\n");
					break;
				}
			case IE_MSG:
				{
					//    
					MsgHider * msg = (MsgHider *)m_InBuf.m_buf;
					MsgAnswHider answer;
					CString cs;

					//       
					memcpy(&answer,msg,sizeof(MsgHider));
					memcpy(answer.v_len,"00000000",8);
					memcpy(answer.h_len,"081",3);
					memcpy(answer.type,"06",2);
					memcpy(answer.code,"00",2);
					CMessage tmpmsg;
					memcpy(answer.rcvr_num,(LPCSTR)tmpmsg.GetKvnString(),12);
					Put(&answer,sizeof(answer));

					//      
					CString msgtext((LPCSTR) &m_InBuf.m_buf[sizeof(MsgHider)],GetMessageSize());
					m_InQueue.Add(CMessage(msgtext));
					TRACE(">>> %.300s\n",msgtext);
					DelIU(); 
					PrintLogWin(m_ncit,"ParseIU():  : %.100s",msgtext);
					PrintLogWin(m_ncit,"ParseIU():     STDP ");
					TRACE("IE_MSG Received\n");
					break;
				}
			case IE_TSTANSW:
				PrintLogWin(m_ncit,"ParseIU():      ");
				TRACE("IE_TSTANSW Answer received\n");
				DelIU(); 
				break;
			default: 
				PrintLogWin(m_ncit,"ParseIU():    GetIUType()=%d",GetIUType());
				TRACE("Unknown IU received GetIUType()=%d\n",GetIUType());
				DelIU(); 
				break;
			}
			break;
		case CONN_HIDER_ERR:
		case CONN_UNK_CLIENT:
		case CONN_UNK_SERVER:
        case CONN_ACCESS:
		case CONN_CORRUPT:
        case CONN_REP_CONNECT:
			PrintLogWin(m_ncit,"ParseIU(): o  STDP  =%d",m_stat);
			PrintLogWin(m_ncit,"ParseIU():    ");
			Reconnect();
			TRACE("ParseUI() CONN_REP_CONNECT\n");
			break;
        case CONN_NOT_SET:
        case CONN_CHAN_CLOSED:
			switch(GetIUType()) {
			case -2: m_stat=CONN_CORRUPT; m_InBuf.Purge(); break;
			case -1: break;
			case IE_CONANSW:
				{
					//    
					ConnAnswHider * ca=(ConnAnswHider *)m_InBuf.m_buf;
					m_stat=(CWinSTDP::WinSTDPStatus)Hex2Int((char *)ca->code,2);
					if(m_stat!=CONN_OK && IsConnected()) {
						PrintLogWin(m_ncit,"ParseIU():     ");
						Disconnect();
						TRACE("ParseUI() CONN_CHAN_CLOSED\n");
					}
					DelIU();
					break;
				}
			default: 
				DelIU(); break;
		}
	}
}

BOOL CWinSTDP::PutMessage(LPCSTR msg)
{
	BOOL retv=FALSE;

	PrintLogWin(m_ncit,"PutMessage():  : %.100s",msg);

	for (int i=0; i<MAX_ACK_WAIT_TIME/500; i++) {
		//    
		if (m_OutQueue.GetSize()==0 && m_OutBuf.Len()==0) {

			m_CrsOut.Lock();
			m_OutQueue.Add(msg);
			m_CrsOut.Unlock();

			//       
			for (int j=0; j<MAX_ACK_WAIT_TIME/500; j++) {
				if (m_OutQueue.GetSize()==0 && m_OutBuf.Len()==0) {
					retv=TRUE;
					break;
				} else Sleep(500);
			}

			break;
		} else Sleep(500);
	}

	if (retv) {
		PrintLogWin(m_ncit,"PutMessage():   ,  ");
	} else {
		Beep(3000,100);
		PrintLogWin(m_ncit,"PutMessage():  ,    ");
	}

	return retv;
}

CString CWinSTDP::GetMessage()
{
	CString rv;
	if(HaveMessage()) {
		m_CrsIn.Lock();
		rv=m_InQueue[0];
		m_InQueue.RemoveAt(0);
		m_CrsIn.Unlock();
	}
	return rv;
}

BOOL CWinSTDP::HaveMessage()
{
	return m_InQueue.GetSize();
}

void CWinSTDP::SendIU()
{
	m_CrsOut.Lock();
	for(int i=0; i<m_OutQueue.GetSize(); i++) {
		CMessage & msg = m_OutQueue[i];
		if(msg.GetSendTime()!=0) continue;
		CString cs=msg;
		Putf("STDP067%08ld0580%02x%s%s%08ld%.12s",
            cs.GetLength(),MSGATR_DEF,m_laa,m_raa,cs.GetLength(),msg.GetKvnString());
		Put((LPCSTR)cs,cs.GetLength());
		msg.SetSendTime();
	}
	m_CrsOut.Unlock();
}

void CWinSTDP::CheckQueue()
{
	m_CrsOut.Lock();
	DWORD cur_time=GetTickCount();
	for(int i=0; i<m_OutQueue.GetSize(); i++) {

		CMessage & msg = m_OutQueue[i];
		if(msg.GetSendTime()==0) continue;
		if(msg.GetAck()) {
			m_OutQueue.RemoveAt(i--);
			continue;
		}
		if(cur_time-msg.GetSendTime()>MAX_ACK_WAIT_TIME && !msg.GetAck()) {
			msg.m_sendtime=0; // try to send this message later
			Reconnect();
			TRACE("CheckQueue()\n");
		}
	}
	m_CrsOut.Unlock();
}

CMessage * CWinSTDP::FindMessage(LPCSTR kvn_str)
{
	m_CrsOut.Lock();
	for(int i=0; i<m_OutQueue.GetSize(); i++) {
		if (m_OutQueue[i].GetKvnString()==CString(kvn_str,12)) {
			m_CrsOut.Unlock();
			return &m_OutQueue[i];
		}
	}
	m_CrsOut.Unlock();
	return NULL;
}

void CWinSTDP::CheckTimeout()
{
	if (GetTickCount()-m_lrecvtime > MAX_RECEIVE_TIMEOUT) {
		PrintLogWin(m_ncit,"CheckTimeout():   ,  ");
		TRACE("CheckTimeout() limit exeeded\n");
		Reconnect();
		m_stat=CONN_NOT_SET;
		m_lrecvtime=GetTickCount();
	}
}

void CWinSTDP::SendTestIU()
{
	if(GetTickCount()-m_lsendtime>SEND_TESTIU_TIMEOUT) {
		if (GetStatus()==WinSTDPStatus::CONN_OK) {
			PrintLogWin(m_ncit,"SendTestIU():    ");
			Putf("STDP0450000000008%s%s00",m_laa,m_raa);
		}
	}
}

CWinSTDP::WinSTDPStatus CWinSTDP::GetStatus()
{
	return m_stat;
}

void CWinSTDP::Reconnect()
{
	AfxBeginThread((AFX_THREADPROC)ReconnectThread,this);
}

void CWinSTDP::PrintLogWin(UINT ncit, LPCSTR fmt, ...)
{

}

void CWinSTDP::ReconnectThread(CWinSTDP &stdp)
{
	stdp.PrintLogWin(stdp.m_ncit,"Reconnect():     (P=%d)",stdp.m_bReconnectInProgress);
	if (!stdp.m_bReconnectInProgress) {
		stdp.m_bReconnectInProgress=TRUE;
		Sleep(1000);
		stdp.Disconnect();
		Sleep(2000);
		stdp.Connect(stdp.m_rip,stdp.m_port,stdp.m_raa,stdp.m_laa,stdp.m_nconf);
		stdp.m_bReconnectInProgress=FALSE;
	}
	stdp.PrintLogWin(stdp.m_ncit,"Reconnect():   ");
}

UINT CWinSTDP::GetOutQueueSize()
{
	return m_OutQueue.GetSize();
}

CString CWinSTDP::GetStatusText()
{
	CString retv="";

	switch (GetStatus()) {
	case CWinSTDP::WinSTDPStatus::CONN_OK: 
		retv="  (CONN_OK)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_HIDER_ERR: 
		retv="   (CONN_HIDER_ERR)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_UNK_CLIENT: 
		retv="   (CONN_UNK_CLIENT)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_UNK_SERVER: 
		retv="   (CONN_UNK_SERVER)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_REP_CONNECT: 
		retv="    (CONN_REP_CONNECT)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_CHAN_CLOSED: 
		retv=" STDP   (CONN_CHAN_CLOSED)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_ACCESS: 
		retv="    (CONN_ACCESS)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_NOT_SET: 
		retv="   (CONN_NOT_SET)";
		break;
	case CWinSTDP::WinSTDPStatus::CONN_CORRUPT: 
		retv="   (CONN_CORRUPT)";
		break;
	default: ;
	}
	return retv;
}
